@@ -1,6 +1,6 @@
 ;;; browse-url.el --- pass a URL to a WWW browser
 
-;; Copyright (C) 1995, 96, 97, 98, 99, 2000, 2001, 2002, 2003
+;; Copyright (C) 1995, 96, 97, 98, 99, 2000, 2001, 2002, 2003, 2005
 ;;   Free Software Foundation, Inc.
 
 ;; Author: Denis Howe <dbh@doc.ic.ac.uk>
@@ -39,7 +39,9 @@
 ;; some of them probably now obsolete:
 
 ;; Function                           Browser     Earliest version
+;; browse-url-seamonkey               SeaMonkey   1st tried with 1.0.4
 ;; browse-url-mozilla                 Mozilla     Don't know
+;; browse-url-firefox                 Firefox     Don't know (1st tried with 1.0.1)
 ;; browse-url-galeon                  Galeon      Don't know
 ;; browse-url-epiphany                Epiphany    Don't know
 ;; browse-url-netscape                Netscape    1.1b1
@@ -56,6 +58,7 @@
 ;; browse-url-generic                 arbitrary
 ;; browse-url-default-windows-browser MS-Windows browser
 ;; browse-url-default-macosx-browser  Mac OS X browser
+;; browse-url-default-xdg-browser     freedesktop.org browser
 ;; browse-url-gnome-moz               GNOME interface to Mozilla
 ;; browse-url-kde                     KDE konqueror (kfm)
 ;; browse-url-opera                   Opera
@@ -248,6 +251,8 @@
    ((memq system-type '(windows-nt ms-dos cygwin32)) ;; XEmacs change
     'browse-url-default-windows-browser)
    ((memq system-type '(darwin)) 'browse-url-default-macosx-browser)
+   ;; XEmacs TODO: add 'browse-url-default-xdg-browser here, but how to detect
+   ;; a freedesktop.org desktop?  executable-find xdg-open is not the way.
    (t 'browse-url-default-browser))
   "*Function to display the current buffer in a WWW browser.
 This is used by the `browse-url-at-point', `browse-url-at-mouse', and
@@ -265,7 +270,9 @@ regexp should probably be \".\" to specify a default browser."
 	  (function-item :tag "Emacs W3M" :value  browse-url-w3)
 	  (function-item :tag "W3M in another Emacs via `gnudoit'"
 			 :value  browse-url-w3-gnudoit)
+	  (function-item :tag "SeaMonkey" :value browse-url-seamonkey)
 	  (function-item :tag "Mozilla" :value  browse-url-mozilla)
+	  (function-item :tag "Firefox" :value browse-url-firefox)
 	  (function-item :tag "Galeon" :value  browse-url-galeon)
 	  (function-item :tag "Epiphany" :value  browse-url-epiphany)
 	  (function-item :tag "Netscape" :value  browse-url-netscape)
@@ -286,6 +293,9 @@ regexp should probably be \".\" to specify a default browser."
 			 :value browse-url-default-windows-browser)
 	  (function-item :tag "Default Mac OS X browser"
 			 :value browse-url-default-macosx-browser)
+	  ;; XEmacs addition
+	  (function-item :tag "Default freedesktop.org browser"
+			 :value browse-url-default-xdg-browser)
 	  (function-item :tag "GNOME invoking Mozilla"
 			 :value browse-url-gnome-moz)
 	  (function-item :tag "Default browser"
@@ -329,6 +339,24 @@ Defaults to the value of `browse-url-netscape-arguments' at the time
   :type '(choice string (const :tag "Default" nil))
   :group 'browse-url)
 
+;;;###autoload
+(defcustom browse-url-seamonkey-program "seamonkey"
+  "*The name by which to invoke SeaMonkey."
+  :type 'string
+  :group 'browse-url)
+
+(defcustom browse-url-seamonkey-arguments nil
+  "*A list of strings to pass to SeaMonkey as arguments."
+  :type '(repeat (string :tag "Argument"))
+  :group 'browse-url)
+
+(defcustom browse-url-seamonkey-startup-arguments browse-url-seamonkey-arguments
+  "*A list of strings to pass to SeaMonkey when it starts up.
+Defaults to the value of `browse-url-seamonkey-arguments' at the time
+`browse-url' is loaded."
+  :type '(repeat (string :tag "Argument"))
+  :group 'browse-url)
+
 ;;;###autoload
 (defcustom browse-url-mozilla-program "mozilla"
   "*The name by which to invoke Mozilla."
@@ -347,6 +375,24 @@ Defaults to the value of `browse-url-mozilla-arguments' at the time
   :type '(repeat (string :tag "Argument"))
   :group 'browse-url)
 
+;;;###autoload
+(defcustom browse-url-firefox-program "firefox"
+  "*The name by which to invoke Firefox."
+  :type 'string
+  :group 'browse-url)
+
+(defcustom browse-url-firefox-arguments nil
+  "*A list of strings to pass to Firefox as arguments."
+  :type '(repeat (string :tag "Argument"))
+  :group 'browse-url)
+
+(defcustom browse-url-firefox-startup-arguments browse-url-firefox-arguments
+  "*A list of strings to pass to Firefox when it starts up.
+Defaults to the value of `browse-url-firefox-arguments' at the time
+`browse-url' is loaded."
+  :type '(repeat (string :tag "Argument"))
+  :group 'browse-url)
+
 ;;;###autoload
 (defcustom browse-url-galeon-program "galeon"
   "*The name by which to invoke Galeon."
@@ -382,6 +428,13 @@ Defaults to the value of `browse-url-epiphany-arguments' at the time
 :type '(repeat (string :tag "Argument"))
 :group 'browse-url)
 
+(defcustom browse-url-seamonkey-new-window-is-tab nil
+  "*Whether to open up new windows in a tab or a new window.
+If non-nil, then open the URL in a new tab rather than a new window if
+`browse-url-seamonkey' is asked to open it in a new window."
+  :type 'boolean
+  :group 'browse-url)
+
 (defcustom browse-url-mozilla-new-window-is-tab nil
   "*Whether to open up new windows in a tab or a new window.
 If non-nil, then open the URL in a new tab rather than a new window if
@@ -389,6 +442,13 @@ If non-nil, then open the URL in a new tab rather than a new window if
   :type 'boolean
   :group 'browse-url)
 
+(defcustom browse-url-firefox-new-window-is-tab nil
+  "*Whether to open up new windows in a tab or a new window.
+If non-nil, then open the URL in a new tab rather than a new window if
+`browse-url-firefox' is asked to open it in a new window."
+  :type 'boolean
+  :group 'browse-url)
+
 (defcustom browse-url-galeon-new-window-is-tab nil
   "*Whether to open up new windows in a tab or a new window.
 If non-nil, then open the URL in a new tab rather than a new window if
@@ -403,6 +463,13 @@ If non-nil, then open the URL in a new tab rather than a new window if
 :type 'boolean
 :group 'browse-url)
 
+(defcustom browse-url-netscape-new-window-is-tab nil
+  "*Whether to open up new windows in a tab or a new window.
+If non-nil, then open the URL in a new tab rather than a new window if
+`browse-url-netscape' is asked to open it in a new window."
+  :type 'boolean
+  :group 'browse-url)
+
 ;;;###autoload
 (defcustom browse-url-new-window-flag nil
   "*If non-nil, always open a new browser window with appropriate browsers.
@@ -798,6 +865,22 @@ to use."
   (interactive (browse-url-interactive-arg "URL: "))
   (start-process (concat "open " url) nil "open" url))
 
+;; XEmacs addition
+;;;###autoload
+(defcustom browse-url-xdg-open-program "xdg-open"
+  "*The name by which to invoke xdg-open."
+  :type 'string
+  :group 'browse-url)
+
+;; XEmacs addition
+;;;###autoload
+(defun browse-url-default-xdg-browser (url &optional new-window)
+  (interactive (browse-url-interactive-arg "URL: "))
+  (let ((process-environment (browse-url-process-environment))
+	(process-connection-type nil))
+    (start-process (concat browse-url-xdg-open-program " " url) nil
+		   browse-url-xdg-open-program url)))
+
 ;; --- Netscape ---
 
 (defun browse-url-process-environment ()
@@ -833,13 +916,18 @@ the effect of `browse-url-new-window-flag'.
 When called non-interactively, optional second argument NEW-WINDOW is
 used instead of `browse-url-new-window-flag'.
 
-The order attempted is emacs-w3m, gnome-moz-remote, Mozilla, Galeon,
-Konqueror, Netscape, Opera, Mosaic, IXI Mosaic, Lynx in an xterm, MMM,
-and then W3."
+The order attempted is freedesktop.org default, gnome-moz-remote, Mozilla,
+Firefox, SeaMonkey, Galeon, Konqueror, Netscape, Opera, Mosaic, IXI Mosaic,
+emacs-w3m, Lynx in an xterm, MMM, and then W3."
   (apply
     (cond
+     ;; XEmacs addition
+     ((executable-find browse-url-xdg-open-program)
+      'browse-url-default-xdg-browser)
      ((executable-find browse-url-gnome-moz-program) 'browse-url-gnome-moz)
      ((executable-find browse-url-mozilla-program) 'browse-url-mozilla)
+     ((executable-find browse-url-firefox-program) 'browse-url-firefox)
+     ((executable-find browse-url-seamonkey-program) 'browse-url-seamonkey)
      ((executable-find browse-url-galeon-program) 'browse-url-galeon)
      ((executable-find browse-url-kde-program) 'browse-url-kde)
      ((executable-find browse-url-netscape-program) 'browse-url-netscape)
@@ -863,6 +951,10 @@ non-nil, load the document in a new Netscape window, otherwise use a
 random existing one.  A non-nil interactive prefix argument reverses
 the effect of `browse-url-new-window-flag'.
 
+If `browse-url-netscape-new-window-is-tab' is non-nil, then
+whenever a document would otherwise be loaded in a new window, it
+is loaded in a new tab in an existing window instead.
+
 When called non-interactively, optional second argument NEW-WINDOW is
 used instead of `browse-url-new-window-flag'."
   (interactive (browse-url-interactive-arg "URL: "))
@@ -872,21 +964,24 @@ used instead of `browse-url-new-window-flag'."
     (setq url (replace-match
 	       (format "%%%x" (string-to-char (match-string 0 url))) t t url)))
   (let* ((process-environment (browse-url-process-environment))
-         (process (apply 'start-process
-			 (concat "netscape " url) nil
-			 browse-url-netscape-program
-			 (append
-			  browse-url-netscape-arguments
-			  (if (eq (console-type) 'mswindows) ;; XEmacs change
-			      (list url)
-			    (append
-			     (if new-window '("-noraise"))
-			     (list "-remote"
-				   (concat "openURL(" url
-					   (if (browse-url-maybe-new-window
-						new-window)
-					       ",new-window")
-					   ")"))))))))
+	 (process
+	  (apply 'start-process
+		 (concat "netscape " url) nil
+		 browse-url-netscape-program
+		 (append
+		  browse-url-netscape-arguments
+		  (if (eq (console-type) 'mswindows) ;; XEmacs change
+		      (list url)
+		    (append
+		     (if new-window '("-noraise"))
+		     (list "-remote"
+			   (concat "openURL(" url
+				   (if (browse-url-maybe-new-window
+					new-window)
+				       (if browse-url-netscape-new-window-is-tab
+					   ",new-tab"
+					 ",new-window"))
+				   ")"))))))))
     (set-process-sentinel process
 			  `(lambda (process change)
 			     (browse-url-netscape-sentinel process ,url)))))
@@ -919,6 +1014,69 @@ How depends on `browse-url-netscape-version'."
            (append browse-url-netscape-arguments
                    (list "-remote" command)))))
 
+;;;###autoload
+(defun browse-url-seamonkey (url &optional new-window)
+  "Ask the SeaMonkey WWW browser to load URL.
+Default to the URL around or before point.  The strings in variable
+`browse-url-seamonkey-arguments' are also passed to SeaMonkey.
+
+When called interactively, if variable `browse-url-new-window-flag' is
+non-nil, load the document in a new SeaMonkey window, otherwise use a
+random existing one.  A non-nil interactive prefix argument reverses
+the effect of `browse-url-new-window-flag'.
+
+If `browse-url-seamonkey-new-window-is-tab' is non-nil, then whenever a
+document would otherwise be loaded in a new window, it is loaded in a
+new tab in an existing window instead.
+
+When called non-interactively, optional second argument NEW-WINDOW is
+used instead of `browse-url-new-window-flag'.
+
+On MS-Windows systems the optional `new-window' parameter is ignored.
+SeaMonkey for Windows does not support the \"-remote\" command line
+parameter. Therefore the `browse-url-new-window-flag', 
+`browse-url-new-window-flag' and `browse-url-seamonkey-new-window-is-tab'
+are ignored as well. SeaMonkey on Windows will always open the
+requested URL in a new window."
+  (interactive (browse-url-interactive-arg "URL: "))
+  ;; URL encode any `confusing' characters in the URL.  This needs to
+  ;; include at least commas; presumably also close parens.
+  (while (string-match "[,)]" url)
+    (setq url (replace-match
+	       (format "%%%x" (string-to-char (match-string 0 url))) t t url)))
+  (let* ((process-environment (browse-url-process-environment))
+         (process
+	  (apply 'start-process
+		 (concat "seamonkey " url) nil
+		 browse-url-seamonkey-program
+		 (append
+		  browse-url-seamonkey-arguments
+          (if (or (featurep 'dos-w32)
+                  (string-match "win32" system-configuration))
+              (list url)
+            (list "-remote"
+                  (concat "openURL("
+                          url
+                          (if (browse-url-maybe-new-window
+                               new-window)
+                              (if browse-url-seamonkey-new-window-is-tab
+                                  ",new-tab"
+                                ",new-window"))
+                          ")")))))))
+    (set-process-sentinel process
+			  `(lambda (process change)
+			     (browse-url-seamonkey-sentinel process ,url)))))
+
+(defun browse-url-seamonkey-sentinel (process url)
+  "Handle a change to the process communicating with SeaMonkey."
+  (or (eq (process-exit-status process) 0)
+      (let* ((process-environment (browse-url-process-environment)))
+	;; SeaMonkey is not running - start it
+	(message "Starting SeaMonkey...")
+	(apply 'start-process (concat "seamonkey " url) nil
+	       browse-url-seamonkey-program
+	       (append browse-url-seamonkey-startup-arguments (list url))))))
+
 ;;;###autoload
 (defun browse-url-mozilla (url &optional new-window)
   "Ask the Mozilla WWW browser to load URL.
@@ -982,6 +1140,69 @@ requested URL in a new window."
 	       browse-url-mozilla-program
 	       (append browse-url-mozilla-startup-arguments (list url))))))
 
+;;;###autoload
+(defun browse-url-firefox (url &optional new-window)
+  "Ask the Firefox WWW browser to load URL.
+Default to the URL around or before point.  The strings in variable
+`browse-url-firefox-arguments' are also passed to Firefox.
+
+When called interactively, if variable `browse-url-new-window-flag' is
+non-nil, load the document in a new Firefox window, otherwise use a
+random existing one.  A non-nil interactive prefix argument reverses
+the effect of `browse-url-new-window-flag'.
+
+If `browse-url-firefox-new-window-is-tab' is non-nil, then whenever a
+document would otherwise be loaded in a new window, it is loaded in a
+new tab in an existing window instead.
+
+When called non-interactively, optional second argument NEW-WINDOW is
+used instead of `browse-url-new-window-flag'.
+
+On MS-Windows systems the optional `new-window' parameter is ignored.
+Firefox for Windows does not support the \"-remote\" command line
+parameter. Therefore the `browse-url-new-window-flag', 
+`browse-url-new-window-flag' and `browse-url-firefox-new-window-is-tab'
+are ignored as well. Firefox on Windows will always open the
+requested URL in a new window."
+  (interactive (browse-url-interactive-arg "URL: "))
+  ;; URL encode any `confusing' characters in the URL.  This needs to
+  ;; include at least commas; presumably also close parens.
+  (while (string-match "[,)]" url)
+    (setq url (replace-match
+	       (format "%%%x" (string-to-char (match-string 0 url))) t t url)))
+  (let* ((process-environment (browse-url-process-environment))
+         (process
+	  (apply 'start-process
+		 (concat "firefox " url) nil
+		 browse-url-firefox-program
+		 (append
+		  browse-url-firefox-arguments
+          (if (or (featurep 'dos-w32)
+                  (string-match "win32" system-configuration))
+              (list url)
+            (list "-remote"
+                  (concat "openURL("
+                          url
+                          (if (browse-url-maybe-new-window
+                               new-window)
+                              (if browse-url-firefox-new-window-is-tab
+                                  ",new-tab"
+                                ",new-window"))
+                          ")")))))))
+    (set-process-sentinel process
+			  `(lambda (process change)
+			     (browse-url-firefox-sentinel process ,url)))))
+
+(defun browse-url-firefox-sentinel (process url)
+  "Handle a change to the process communicating with Firefox."
+  (or (eq (process-exit-status process) 0)
+      (let* ((process-environment (browse-url-process-environment)))
+	;; Firefox is not running - start it
+	(message "Starting Firefox...")
+	(apply 'start-process (concat "firefox " url) nil
+	       browse-url-firefox-program
+	       (append browse-url-firefox-startup-arguments (list url))))))
+
 ;;;###autoload
 (defun browse-url-galeon (url &optional new-window)
   "Ask the Galeon WWW browser to load URL.
@@ -1133,28 +1354,26 @@ used instead of `browse-url-new-window-flag'."
   (let ((pidfile (expand-file-name browse-url-mosaic-pidfile))
 	pid)
     (if (file-readable-p pidfile)
-	(save-excursion
-	  (find-file pidfile)
-	  (goto-char (point-min))
-	  (setq pid (read (current-buffer)))
-	  (kill-buffer nil)))
-    (if (and pid (zerop (signal-process pid 0))) ; Mosaic running
-	(save-excursion
-	  (find-file (format "/tmp/Mosaic.%d" pid))
-	  (erase-buffer)
-	  (insert (if (browse-url-maybe-new-window new-window)
-		      "newwin\n"
-		    "goto\n")
-		  url "\n")
-	  (save-buffer)
-	  (kill-buffer nil)
+        (with-temp-buffer
+          (insert-file-contents pidfile)
+	  (setq pid (read (current-buffer)))))
+    (if (and (integerp pid) (zerop (signal-process pid 0))) ; Mosaic running
+        (progn
+          (with-temp-buffer
+            (insert (if (browse-url-maybe-new-window new-window)
+                        "newwin\n"
+                      "goto\n")
+                    url "\n")
+            (if (file-exists-p (setq pidfile (format "/tmp/Mosaic.%d" pid)))
+                (delete-file pidfile))
+            ;; http://debbugs.gnu.org/17428.  Use O_EXCL.
+            (write-region nil nil pidfile nil 'silent nil 'excl))
 	  ;; Send signal SIGUSR to Mosaic
 	  (message "Signalling Mosaic...")
 	  (signal-process pid 'SIGUSR1)
 	  ;; Or you could try:
 	  ;; (call-process "kill" nil 0 nil "-USR1" (int-to-string pid))
-	  (message "Signalling Mosaic...done")
-	  )
+	  (message "Signalling Mosaic...done"))
       ;; Mosaic not running - start it
       (message "Starting Mosaic...")
       (apply 'start-process "xmosaic" nil browse-url-mosaic-program
